"""Create file-like object from iterable/iterator."""
import io
from array import array
from typing import Iterable, Iterator, Optional, Union


class StreamFromIter(io.RawIOBase):
    """Stream bytes from iterable/iterator.

    >>> def chunks():
    ...     for chunk in [b"foo", b"bar", b"spam"]:
    ...         yield chunk
    >>> with io.BufferedReader(StreamFromIter(chunks())) as stream:
    ...     print(stream.read())
    b'foobarspam'
    >>> with StreamFromIter([b"foo", b"bar", b"spam"]) as stream:
    ...     print(stream.read())
    b'foobarspam'
    """

    def __init__(
        self,
        source: Union[  # type: ignore
            Iterator[Union[bytearray, array]],
            Iterable[Union[bytearray, array]],
        ],
    ):
        """Implement Python io stream protocol.

        Args:
            source: iterable or iterator that produce bytes.
        """
        self.leftover: Optional[Union[bytearray, array]] = None  # type: ignore
        self.iterator = source if isinstance(source, Iterator) else iter(source)
        super().__init__()

    def readable(self) -> bool:
        """Implement abstract method in io.IOBase.

        Return:
             const True to confirm the stream can be read from.
        """
        return True

    def readinto(self, allocated_buffer: Union[bytearray, array]) -> int:  # type: ignore
        """Implement abstract method in io.RawIOBase.

        Read bytes into a pre-allocated, writable bytes-like object b,
        and return the number of bytes read.
        For example, b might be a bytearray.
        If the object is in non-blocking mode and no bytes are available, None is returned.

        Args:
            allocated_buffer: buffer to read into.

        Return:
            Number of read bytes.
        """
        try:
            max_bytes_in_result = len(allocated_buffer)
            chunk = self.leftover or next(self.iterator)
            output, self.leftover = chunk[:max_bytes_in_result], chunk[max_bytes_in_result:]
            allocated_buffer[: len(output)] = output  # type: ignore
            return len(output)
        except StopIteration:
            return 0  # indicate EOF


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
pytest fixture configurations
"""
import os
import random
import string
from unittest.mock import patch

import aiobotocore
import aiobotocore.session
import botocore.stub
import pytest
import pytest_asyncio

MAX_FILE_LENGTH = 16
MAX_FILE_CHUNK_NUMBER = 4
S3_NAME_ALPHABET = string.ascii_letters + string.digits + "-_"
FILE_BODY_ALPHABET = (string.ascii_letters + string.digits + "-_").encode()
S3_BUCKET_NAME_LENGTH = 12
S3_FILE_FOLDER_LENGTH = 12
S3_FILE_NAME_LENGTH = 12


@pytest_asyncio.fixture(scope="function")
async def s3_stub() -> botocore.stub.Stubber:
    # to test aiobotocore we cannot use moto
    with patch.dict(os.environ, {"AWS_ACCESS_KEY_ID": "fake", "AWS_SECRET_ACCESS_KEY": "fake"}):
        session = aiobotocore.session.get_session()
        async with session.create_client("s3") as stubbed_client:
            with botocore.stub.Stubber(stubbed_client) as stubber:
                yield stubber
                stubber.assert_no_pending_responses()


@pytest.fixture(scope="function")
def file_content() -> bytes:
    return b"".join(
        random.choice(FILE_BODY_ALPHABET).to_bytes(1, "big")
        for _ in range(random.randint(0, MAX_FILE_LENGTH))
    )


@pytest.fixture(scope="function")
def chunk_size() -> int:
    return random.randint(1, MAX_FILE_LENGTH)


@pytest.fixture(scope="function")
def bucket() -> str:
    return "".join(random.choice(S3_NAME_ALPHABET) for _ in range(S3_BUCKET_NAME_LENGTH))


@pytest.fixture(scope="function")
def s3_file_name() -> str:
    return "".join(random.choice(S3_NAME_ALPHABET) for _ in range(S3_FILE_NAME_LENGTH))


@pytest.fixture(scope="function")
def s3_folder() -> str:
    return "".join(random.choice(S3_NAME_ALPHABET) for _ in range(S3_FILE_FOLDER_LENGTH))



import os.path

import botocore.stub
import pytest

from aios3.file import chunks, read, save, stream


class Stream:
    def __init__(self, file_content):
        self.file_content = file_content

    async def read(self, amt):
        if self.file_content is None:
            return b""
        if amt is None:
            result = self.file_content
            self.file_content = None
            return result
        result, self.file_content = self.file_content[:amt], self.file_content[amt:]
        return result

    def close(self):
        pass


@pytest.mark.asyncio
async def test_aios3_file_save(s3_stub: botocore.stub.Stubber, bucket, s3_file_name):
    body = b"123"
    folder = "/456"
    key = os.path.join(folder, s3_file_name)
    s3_stub.add_response(
        "put_object", {}, expected_params={"Body": body, "Bucket": bucket, "Key": key}
    )
    await save(s3=s3_stub.client, bucket=bucket, key=key, body=body)


@pytest.mark.asyncio
async def test_aios3_file_chunks(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content, chunk_size
):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert [
        chunk async for chunk in chunks(s3=s3_stub.client, bucket=bucket, key=key, amt=chunk_size)
    ] == [
        file_content[chunk_idx : chunk_idx + chunk_size]
        for chunk_idx in range(0, len(file_content), chunk_size)
    ]


@pytest.mark.asyncio
async def test_aios3_file_stream(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content, chunk_size
):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert (
        await stream(s3=s3_stub.client, bucket=bucket, key=key, amt=chunk_size)
    ).read() == file_content


@pytest.mark.asyncio
async def test_aios3_file_chunks_default_chunk_size(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content
):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert [
        chunk
        async for chunk in chunks(
            s3=s3_stub.client,
            bucket=bucket,
            key=key,
        )
    ] == ([file_content] if file_content else [])


@pytest.mark.asyncio
async def test_aios3_file_read_default_chunk_size(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content
):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert (
        await read(
            s3=s3_stub.client,
            bucket=bucket,
            key=key,
        )
        == file_content
    )


@pytest.mark.asyncio
async def test_aios3_file_stream_default_chunk_size(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content
):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert (
        await stream(
            s3=s3_stub.client,
            bucket=bucket,
            key=key,
        )
    ).read() == file_content


@pytest.mark.asyncio
async def test_aios3_file_read(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, file_content, chunk_size
):
    await check_reading(bucket, chunk_size, file_content, s3_file_name, s3_folder, s3_stub)


@pytest.mark.asyncio
async def test_aios3_file_read_zero_length(
    s3_stub: botocore.stub.Stubber, bucket, s3_file_name, s3_folder, chunk_size
):
    file_content = b""
    await check_reading(bucket, chunk_size, file_content, s3_file_name, s3_folder, s3_stub)


async def check_reading(bucket, chunk_size, file_content, s3_file_name, s3_folder, s3_stub):
    key = os.path.join(s3_folder, s3_file_name)
    s3_stub.add_response(
        "get_object",
        {"Body": Stream(file_content)},
        expected_params={"Bucket": bucket, "Key": key},
    )
    assert await read(s3=s3_stub.client, bucket=bucket, key=key, amt=chunk_size) == file_content



"""Include requirements.txt in pyproject.toml project.dependencies array."""
import toml

PROJECT_METADATA_FILE_NAME = "pyproject.toml"
REQUIREMENTS_FILE_NAME = "requirements.txt"


with open(REQUIREMENTS_FILE_NAME, "r", encoding="utf8") as f:
    requirements = [
        line.strip() for line in f if line.strip() and not line.lstrip().startswith("#")
    ]
print(
    f"From {REQUIREMENTS_FILE_NAME} read requirements:"
    f"\n{requirements[:10]}...\n{len(requirements)} total"
)

with open(PROJECT_METADATA_FILE_NAME, "r", encoding="utf8") as f:
    pyproject_data = toml.load(f)

print(
    f"Updating project {pyproject_data['project']['name']} project.dependencies array"
    f" in {PROJECT_METADATA_FILE_NAME}"
)

pyproject_data["project"]["dependencies"] = requirements

with open(PROJECT_METADATA_FILE_NAME, "w", encoding="utf8") as f:
    toml.dump(pyproject_data, f)



"""S3 "files" operations with aiobotocore."""
import contextlib
import io
from typing import IO, AsyncGenerator, Optional

from aiobotocore.session import AioBaseClient, get_session

from aios3.stream_iter import StreamFromIter


async def save(  # pylint: disable= invalid-name
    bucket: str, key: str, body: bytes, s3: Optional[AioBaseClient] = None
) -> None:
    """Create the file with the `body`.

    Args:
        bucket: S3 bucket.
        key: path in the bucket including "file name".
        body: content to write into file.
        s3: boto s3 client. by default it auto-created inside the function.
    """
    session = get_session() if s3 is None else None
    async with session.create_client(
        "s3"
    ) if s3 is None else contextlib.AsyncExitStack() as s3_temp:
        await (s3 or s3_temp).put_object(Body=body, Bucket=bucket, Key=key)


async def read(  # pylint: disable= invalid-name
    bucket: str, key: str, amt: Optional[int] = None, s3: Optional[AioBaseClient] = None
) -> bytes:
    """Read the full content of the file as bytes.

    Args:
        bucket: S3 bucket.
        key: path in the bucket including "file name".
        amt: max number of bytes to read in one chunk. by default read all.
        s3: boto s3 client. by default it auto-created inside the function.

    Return:
        The file content as bytes.
    """
    return b"".join([chunk async for chunk in chunks(bucket=bucket, key=key, amt=amt, s3=s3)])


async def chunks(  # pylint: disable= invalid-name
    bucket: str, key: str, amt: Optional[int] = None, s3: Optional[AioBaseClient] = None
) -> AsyncGenerator[bytearray, None]:
    """Generate file chunks `amt` bytes max.

    Args:
        bucket: S3 bucket.
        key: path in the bucket including "file name".
        amt: max number of bytes to read in one chunk. by default read file as one chunk.
        s3: boto s3 client. by default it auto-created inside the function.

    Return:
        Chunks of the file.
    """
    session = get_session() if s3 is None else None
    async with session.create_client(
        "s3"
    ) if s3 is None else contextlib.AsyncExitStack() as s3_temp:
        s3_file_obj = await (s3 or s3_temp).get_object(Bucket=bucket, Key=key)
        s3_file_stream = s3_file_obj["Body"]
        chunks_count = 0
        try:
            while True:
                chunk = await s3_file_stream.read(amt)
                chunks_count += 1
                if len(chunk) == 0:
                    break
                yield chunk
        finally:
            s3_file_stream.close()


async def stream(  # pylint: disable= invalid-name
    bucket: str, key: str, amt: Optional[int] = None, s3: Optional[AioBaseClient] = None
) -> IO[bytes]:
    """Create file-like object to stream the file content.

    Args:
        bucket: S3 bucket.
        key: path in the bucket including "file name".
        amt: max number of bytes to read in one chunk. by default read file as one chunk.
        s3: boto s3 client. by default it auto-created inside the function.

    Return:
        Python file stream with the file content.
    """
    file_chunks = [chunk async for chunk in chunks(bucket=bucket, key=key, amt=amt, s3=s3)]
    return io.BufferedReader(StreamFromIter(file_chunks))
